from toolmate import config, convertOutputText
from toolmate.utils.call_llm import CallLLM
from toolmate.gui.worker import QtApiResponseStreamer
from toolmate.utils.tts_utils import TTSUtil
import getpass, requests, json

from PySide6.QtCore import Qt, QThreadPool
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QCompleter, QMainWindow, QWidget, QMessageBox, QPlainTextEdit, QProgressBar, QHBoxLayout, QVBoxLayout, QLineEdit, QSplitter, QComboBox, QPushButton

class CentralWidget(QWidget):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        # set up interface
        self.setupUI()
        # set up variables
        self.setupVariables()
    
    def setupVariables(self):
        # username
        self.user = getpass.getuser().split(" ")[0].capitalize()
        self.assistant = config.toolMateAIName.split(" ")[0]
        # load LetMeDoIt AI system message
        config.currentMessages=CallLLM.resetMessages()
        # thread pool
        self.threadpool = QThreadPool()
        # scroll bar
        self.contentScrollBar = self.contentView.verticalScrollBar()
        # operations
        self.newSession = True
        self.lastResponse = ""

    def setupUI(self):
        # a layout with left and right columns and a splitter placed between them
        layout000 = QHBoxLayout()
        self.setLayout(layout000)
        
        widgetLt = QWidget()
        layout000Lt = QVBoxLayout()
        widgetLt.setLayout(layout000Lt)

        widgetRt = QWidget()
        layout000Rt = QVBoxLayout()
        rtTop = QWidget()
        rtTopLayout = QVBoxLayout()
        rtTop.setLayout(rtTopLayout)
        rtBottom = QWidget()
        rtBottomLayout = QVBoxLayout()
        rtBottom.setLayout(rtBottomLayout)
        rtSplitter = QSplitter(Qt.Vertical, self)
        rtSplitter.addWidget(rtTop)
        rtSplitter.addWidget(rtBottom)
        layout000Rt.addWidget(rtSplitter)

        widgetRt.setLayout(layout000Rt)
        
        splitter = QSplitter(Qt.Horizontal, self)
        splitter.addWidget(widgetLt)
        splitter.addWidget(widgetRt)
        layout000.addWidget(splitter)

        # add widgets to left column later
        widgetLt.hide()

        # widgets
        # user input
        self.userInput = QLineEdit()
        self.userInputMultiline = QPlainTextEdit()
        self.userInputMultiline.setPlaceholderText("Enter your request here ...")
        completer = QCompleter([f"@{i}" for i in self.getAllTools()]+["tm -exec", "tmc -exec"])
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        completer.setFilterMode(Qt.MatchContains)
        self.userInput.setCompleter(completer)
        self.userInput.setPlaceholderText("Find a tool here ...")
        self.userInput.mousePressEvent = lambda _ : self.userInput.selectAll()
        self.userInput.setClearButtonEnabled(True)
        self.sendButton = QPushButton("Send")
        self.sendButton.clicked.connect(self.submit)
        # content view
        self.contentView = QPlainTextEdit()
        self.contentView.setReadOnly(True)
        self.setFontSize()
        # progress bar
        self.progressBar = QProgressBar()
        self.progressBar.setRange(0, 0) # Set the progress bar to use an indeterminate progress indicator
        
        # update layout
        rtTopLayout.addWidget(self.contentView)
        rtBottomLayout.addWidget(self.userInput)
        rtBottomLayout.addWidget(self.userInputMultiline)
        rtBottomLayout.addWidget(self.sendButton)
        rtBottomLayout.addWidget(self.progressBar)
        self.progressBar.hide()

        # Connections
        self.userInput.returnPressed.connect(self.addTool)

    def getAllTools(self):
        query = "@"
        endpoint = f"{config.toolmate_api_client_host}:{config.toolmate_api_client_port}/api/tools"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
            results = json.loads(response.json())["results"]
            return list(results.keys())
        except Exception as e:
            #response = f"Error: {e}"
            pass
        return []

    def setFontSize(self, index=None):
        if index is not None:
            config.desktopAssistantFontSize = index + 1
        # content view
        font = self.contentView.font()
        font.setPointSize(config.desktopAssistantFontSize)
        self.contentView.setFont(font)
        config.saveConfig()

    def addTool(self):
        toolText = self.userInput.text().strip()
        if toolText:
            self.userInputMultiline.insertPlainText(f" {toolText} ")
            self.userInput.setText("")

    def submit(self):
        toolText = self.userInput.text().strip()
        execute = (toolText in ("tm -exec", "tmc -exec"))
        if execute:
            request = toolText
        else:
            requestText = self.userInputMultiline.toPlainText().strip()
            request = f"{toolText} {requestText}".strip()
        if request:
            self.userInput.setDisabled(True)
            self.addContent(request)

            if request in ("tm -exec", "tmc -exec"):
                request = "@command" if "```command" in self.lastResponse else "@execute_python_code"

            self.sendButton.hide()
            self.progressBar.show()
            QtApiResponseStreamer(self).workOnRequest(request, chat=False if self.newSession else True)
            self.newSession = False

            if not execute:
                self.userInputMultiline.setPlainText("")

    def processResponse(self):
        self.userInput.setText("")
        self.userInput.setEnabled(True)
        self.progressBar.hide()
        self.sendButton.show()
        self.userInputMultiline.setFocus()
        # handle user confirmation
        if "Run `tm -exec` or `tmc -exec` to confirm!" in self.lastResponse and self.confirmCodeExecution():
            self.userInput.setText("tm -exec")
            self.submit()

    def confirmCodeExecution(self):
        msgBox = QMessageBox(QMessageBox.Warning,
                             "User confirmation required!",
                             "Do you want to execute the task?",
                             QMessageBox.NoButton, self)
        msgBox.addButton("No", QMessageBox.RejectRole)
        msgBox.addButton("Yes", QMessageBox.AcceptRole)
        answer = msgBox.exec_()
        if answer and not answer == 2:
            return True
        else:
            # Continue
            return False

    def streamResponse(self, apiResponse):
        # in case of group discussion; getting the last message only is not enough
        conversation = json.loads(apiResponse)
        messages = []
        for i in conversation:
            role = i.get("role", "")
            content = i.get("content", "")
            if role in ("user", "assistant"):
                if role == "assistant":
                    content = convertOutputText(content.rstrip())
                    messages.append(f"[{self.assistant}] {content}")
                    self.lastResponse = content
                else:
                    messages.append(f"[{self.user}] {content}")
        self.contentView.setPlainText("\n\n".join(messages))
        self.contentScrollBar.setValue(self.contentScrollBar.maximum())

        #if config.ttsOutput:
        #    TTSUtil.play("")

    def addContent(self, newContent, user=True) -> None:
        content = self.contentView.toPlainText()
        if content:
            self.contentView.insertPlainText("\n\n")
        if user:
            self.contentView.insertPlainText(f"[{self.user}] {newContent}\n\n[{self.assistant}] ")
        else:
            self.contentView.insertPlainText(newContent)

class DesktopAssistant(QMainWindow):
    def __init__(self, standalone=False) -> None:
        super().__init__()
        # check if running standalone
        self.standalone = standalone
        # set title
        self.setWindowTitle(config.toolMateAIName)
        # gui
        self.initUI()
        # shortcuts
        self.processResponse = self.centralWidget.processResponse
        self.streamResponse = self.centralWidget.streamResponse
        self.setFontSize = self.centralWidget.setFontSize

    def closeEvent(self, event):
        if self.standalone:
            event.accept()
        else:
            # hiding it, instead of closing it, to save from reloading time
            event.ignore()
            self.hide()

    def initUI(self):
        config.conversationStarted = False
        self.centralWidget = CentralWidget(self)
        self.setCentralWidget(self.centralWidget)
        self.resize(config.desktopAssistantWidth, config.desktopAssistantHeight)
        #
        self.createMenubar()

    def resizeEvent(self, event):
        size = event.size()
        config.desktopAssistantWidth = size.width()
        config.desktopAssistantHeight = size.height()
        config.saveConfig()
    
    def printTextOutput(self, output):
        self.centralWidget.addContent(output, False)

    def showFontSizeComboBox(self):
        self.fontSizeComboBox = QComboBox()
        self.fontSizeComboBox.setWindowTitle("Select Font Size")
        self.fontSizeComboBox.setFixedWidth(400)
        self.fontSizeComboBox.addItems([str(i) for i in range(1, 51)])
        self.fontSizeComboBox.setCurrentIndex((config.desktopAssistantFontSize - 1))
        self.fontSizeComboBox.currentIndexChanged.connect(self.setFontSize)
        self.fontSizeComboBox.show()

    def toggleAutoPaste(self):
        config.pasteTextOnWindowActivation = not config.pasteTextOnWindowActivation
        config.saveConfig()

    def createMenubar(self):
        # Create a menu bar
        menubar = self.menuBar()

        # Create a File menu and add it to the menu bar
        file_menu = menubar.addMenu("ToolMate AI")

        new_action = QAction("New Session", self)
        new_action.setShortcut("Ctrl+N")
        new_action.triggered.connect(self.newConversation)
        file_menu.addAction(new_action)

        file_menu.addSeparator()

        new_action = QAction("Send", self)
        new_action.setShortcut("Ctrl+S")
        new_action.triggered.connect(self.centralWidget.submit)
        file_menu.addAction(new_action)

        file_menu.addSeparator()

        new_action = QAction("Change Font Size", self)
        new_action.triggered.connect(self.showFontSizeComboBox)
        file_menu.addAction(new_action)

        file_menu.addSeparator()

        new_action = QAction("Toggle Auto Paste", self)
        new_action.triggered.connect(self.toggleAutoPaste)
        file_menu.addAction(new_action)

        file_menu.addSeparator()

        new_action = QAction("Close", self)
        new_action.triggered.connect(self.hide)
        file_menu.addAction(new_action)

        """new_action = QAction("Open", self)
        new_action.setShortcut("Ctrl+O")
        new_action.triggered.connect(self.openConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Save", self)
        new_action.setShortcut("Ctrl+S")
        new_action.triggered.connect(self.saveConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Save As...", self)
        new_action.triggered.connect(self.saveAsConversation)
        file_menu.addAction(new_action)

        new_action = QAction("Export", self)
        new_action.triggered.connect(self.exportConversation)
        file_menu.addAction(new_action)

        file_menu.addSeparator()"""

    def newConversation(self):
        self.centralWidget.contentView.setPlainText("")
        self.centralWidget.newSession = True


from toolmate import config, fineTunePythonCode, displayPythonCode
from toolmate import print2, print3
import traceback
from toolmate import installPipPackage
import io, sys
from io import StringIO

"""
ToolMate AI Plugin - auto correct python code

functionalities:
* install missing packages
* fixed broken codes

User can define the maximum number of auto-correctioning attempts by editing "max_consecutive_auto_correction" in config.py.
The default value of config.max_consecutive_auto_correction is 3.

[TOOL_CALL]
"""

def correct_python_code(function_args):

    # Create a StringIO object to capture the output
    thisOutput = StringIO()
    # Redirect stdout to the StringIO object
    old_stdout = sys.stdout
    sys.stdout = thisOutput

    # get the sql query statement
    issue = function_args.get("issue") # required
    print(config.divider)
    print(f"# Issue\n{issue}")

    fix = function_args.get("corrected_code") # required
    missing = function_args.get("missing_module", "") # required
    if missing in ("[]", "???"):
        missing = ""

    try:
        if missing:
            try:
                message = f"Installing missing package: {missing}"
                print(message) if hasattr(config, "api_server_id") else print3(message)
                installPipPackage(f"--upgrade {missing}")
            except:
                print(traceback.format_exc())
        if config.developer or config.codeDisplay:
            print("# Improved code")
            displayPythonCode(fix)
        print("Running improved code ...")
        print(config.divider)
        exec(fineTunePythonCode(fix), globals())
        if config.pythonFunctionResponse:
            print(config.pythonFunctionResponse)

        # Restore the original stdout
        sys.stdout = old_stdout

        return f"[EXECUTED]{thisOutput.getvalue().strip()}"
    except:
        # Restore the original stdout
        sys.stdout = old_stdout

        return traceback.format_exc()

functionSignature = {
    "examples": [
        "fix bug",
        "correct code",
    ],
    "name": "correct_python_code",
    "description": "Fix Python code if both the original code and the traceback error are provided",
    "parameters": {
        "type": "object",
        "properties": {
            "corrected_code": {
                "type": "string",
                "description": "Generate an improved version of python code that resolved the traceback error. Return the original code only if traceback shows an import error.",
            },
            "missing_module": {
                "type": "string",
                "description": """The module name identified in ModuleNotFoundError, if any. Return '' if there is no import error in the traceback.""",
            },
            "issue": {
                "type": "string",
                "description": """Briefly explain the error""",
            },
        },
        "required": ["corrected_code", "missing_module", "issue"],
    },
}

# configs particular to this plugin
# persistent
persistentConfigs = (
    ("max_consecutive_auto_correction", 5),
)
config.setConfig(persistentConfigs)

config.addFunctionCall(signature=functionSignature, method=correct_python_code)


import os, re, subprocess
from toolmate import config
localStorage = config.localStorage
if os.getcwd() != localStorage:
    os.chdir(localStorage)

try:
    from PySide6.QtWidgets import QSystemTrayIcon, QMenu, QApplication, QMessageBox
    from PySide6.QtGui import QIcon, QAction, QGuiApplication
    from PySide6.QtCore import QEvent
except:
    print("Module 'gui' not found! Run 'pip install toolmate[gui]' first!")
    exit()

import sys, platform, webbrowser, shutil, pyperclip
from toolmate import startAutogenstudioServer, runToolMateCommand, isServerAlive, print2, getCliOutput
from toolmate.gui.desktop_assistant import DesktopAssistant
from pathlib import Path
from functools import partial
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from toolmate.utils.tts_utils import TTSUtil

toolMateAIFile = os.path.realpath(__file__)
toolMateAIFolder = os.path.dirname(toolMateAIFile)
#with open(os.path.join(toolMateAIFolder, "package_name.txt"), "r", encoding="utf-8") as fileObj:
#    package = fileObj.read()
iconFile = os.path.join(toolMateAIFolder, "icons", "ai.png")
thisOS = platform.system()


class ToolMateHub(QSystemTrayIcon):

    def showDesktopAssistant(self):
        # to work with mutliple virtual desktops
        config.desktopAssistant.hide()
        config.desktopAssistant.show()
        config.desktopAssistant.centralWidget.userInputMultiline.setFocus()

    def startApiServer(self):
        host = config.toolmate_api_client_host
        port = config.toolmate_api_client_port
        if not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
            configFile = os.path.join(config.toolMateAIFolder, "config.py")
            if (os.path.getsize(configFile) == 0 or not hasattr(config, "llmInterface") or not config.llmInterface) and shutil.which("tmsetup"):
                os.system("tmsetup")
            if shutil.which("toolmateserver"):
                print("Loading ToolMate AI ...")
                if shutil.which("nohup"):
                    cli = f'''{shutil.which("nohup")} "{shutil.which("toolmateserver")}" > ~/toolmate/nohup-api-server.out 2>&1 &'''
                    os.system(cli)
                else:
                    subprocess.Popen(shutil.which("toolmateserver"), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                # wait until the server is up
                while not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
                    pass
                print("ToolMate AI server is now running ...")
            else:
                #print("Failed to connect ToolMate AI! Run `toolmateserver` first!")
                QMessageBox.information(self, "ToolMate AI", "Failed to connect ToolMate AI! Run `toolmateserver` first!")

    def __init__(self, icon, parent=None):
        super().__init__(icon, parent)

        # start API server if it is not running
        self.startApiServer()

        # pre-load desktop assistant gui
        config.desktopAssistant = DesktopAssistant()

        self.clipboard = PyperclipClipboard()
        self.menu = QMenu(parent)

        #if config.developer:
        chatgui = QAction("Desktop Assistant", self)
        chatgui.triggered.connect(self.showDesktopAssistant)
        self.menu.addAction(chatgui)
        self.menu.addSeparator()

        action = QAction("ToolMate Terminal", self)
        action.triggered.connect(partial(runToolMateCommand, "toolmate"))
        self.menu.addAction(action)
        self.menu.addSeparator()

        """for i in ("toolmate", "letmedoit"):
            action = QAction(i, self)
            action.triggered.connect(partial(runToolMateCommand, i))
            self.menu.addAction(action)
        self.menu.addSeparator()

        # submenu - servers
        submenu = QMenu()
        servers = [
            "toolserver",
            "chatserver",
            "visionserver",
        ]
        if config.customToolServer_command:
            servers.append("customtoolserver")
        if config.customChatServer_command:
            servers.append("customchatserver")
        if config.customVisionServer_command:
            servers.append("customvisionserver")
        for i in servers:
            action = QAction(i, self)
            action.triggered.connect(partial(runToolMateCommand, i))
            submenu.addAction(action)

        menuAction = QAction("Servers", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)

        # submenu - cahtbots
        submenu = QMenu()
        for i in (
            "llamacpppython",
            "llamacppserver",
            "ollamachat",
            "groqchat",
            "openai",
            "geminipro",
            "geminiprovision",
            "palm2",
            "codey",
        ):
            action = QAction(i[:-4] if i.endswith("chat") else i, self)
            action.triggered.connect(partial(runToolMateCommand, i))
            submenu.addAction(action)

        menuAction = QAction("Chatbots", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)"""

        # submenu - autogen agents
        submenu = QMenu()
        for i in ("autoassist", "autoretrieve", "autobuild", "autocaptain"):
            action = QAction(i, self)
            action.triggered.connect(partial(runToolMateCommand, i))
            submenu.addAction(action)

        # autogenstudio
        action = QAction("autogenstudio", self)
        action.triggered.connect(startAutogenstudioServer)
        submenu.addAction(action)

        menuAction = QAction("AutoGen Agents", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)

        # submenu - researches
        submenu = QMenu()

        action = QAction("NotebookLM", self)
        action.triggered.connect(lambda: webbrowser.open("https://notebooklm.google.com/"))
        submenu.addAction(action)

        action = QAction("Perplexica", self)
        action.triggered.connect(lambda: webbrowser.open("http://localhost:3000"))
        submenu.addAction(action)

        action = QAction("SearXNG", self)
        action.triggered.connect(lambda: webbrowser.open("http://localhost:4000"))
        submenu.addAction(action)

        # submenu - clipboard
        """submenu = QMenu()

        action = QAction("read", self)
        action.triggered.connect(self.readClipboard)
        submenu.addAction(action)

        action = QAction("prompt", self)
        action.triggered.connect(lambda: runToolMateCommand("toolmate -p true -u false -n true -i false"))
        submenu.addAction(action)

        action = QAction("summarize", self)
        action.triggered.connect(lambda: runToolMateCommand("toolmate -rp true -u false -n true -i false -c 'Let me Summarize'"))
        submenu.addAction(action)

        action = QAction("explain", self)
        action.triggered.connect(lambda: runToolMateCommand("toolmate -rp true -u false -n true -i false -c 'Let me Explain'"))
        submenu.addAction(action)

        action = QAction("translate", self)
        action.triggered.connect(lambda: runToolMateCommand("toolmate -rp true -u false -n true -i false -c 'Let me Translate'"))
        submenu.addAction(action)

        action = QAction("edit", self)
        action.triggered.connect(lambda: runToolMateCommand("etextedit -p true"))
        submenu.addAction(action)

        menuAction = QAction("Clipboard", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)

        # submenu - utilities
        submenu = QMenu()

        action = QAction("perplexica", self)
        action.triggered.connect(self.launchPerplexica)
        submenu.addAction(action)

        if hasattr(config, "searx_tabs"):
            action = QAction("searxng", self)
            action.triggered.connect(lambda: webbrowser.open(f"http://{config.searx_server}:{config.searx_port}"))
            submenu.addAction(action)

        for i in (
            #"rag",
            "etextedit",
            "commandprompt",
        ):
            action = QAction(i, self)
            action.triggered.connect(partial(runToolMateCommand, i))
            submenu.addAction(action)

        menuAction = QAction("Utilities", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)

        # submenu - custom commands
        if hasattr(config, "customTrayCommands") and config.customTrayCommands:
            submenu = QMenu()
            for i in config.customTrayCommands:
                if not i:
                    self.menu.addSeparator()
                else:
                    action = QAction("ollama" if i == "ollamachat" else i, self)
                    action.triggered.connect(partial(runToolMateCommand, i))
                    submenu.addAction(action)

        menuAction = QAction("Custom", self)
        menuAction.setMenu(submenu)
        self.menu.addAction(menuAction)"""

        self.menu.addSeparator()

        action = QAction("Edit Configurations", self)
        action.triggered.connect(partial(runToolMateCommand, "tmsetup -m"))
        self.menu.addAction(action)
        self.menu.addSeparator()

        helpAction = QAction("Wiki", self)
        helpAction.triggered.connect(lambda: webbrowser.open("https://github.com/eliranwong/letmedoit/wiki"))
        self.menu.addAction(helpAction)

        self.menu.addSeparator()

        exitAction = QAction("Exit", self)
        exitAction.triggered.connect(self.exit)
        self.menu.addAction(exitAction)

        self.setContextMenu(self.menu)

        # show desktop assistant on startup
        config.desktopAssistant.show()
        config.desktopAssistant.centralWidget.userInputMultiline.setFocus()

    def exit(self):
        self.setVisible(False)
        QGuiApplication.instance().quit()

    def readClipboard(self):
        if config.terminalEnableTermuxAPI:
            clipboardText = getCliOutput("termux-clipboard-get")
        else:
            clipboardText = self.clipboard.get_data().text
        TTSUtil.play(re.sub(config.tts_doNotReadPattern, "", clipboardText))

    def launchPerplexica(self):
        current_dir = os.getcwd()
        if config.perplexica_directory and not os.path.isdir(config.perplexica_directory):
            config.perplexica_directory = ""
            config.saveConfig()
        if not config.perplexica_directory:
            perplexica_directory = os.path.join(config.localStorage, "Perplexica")
            if os.path.isdir(perplexica_directory):
                config.perplexica_directory = perplexica_directory
                config.saveConfig()
            elif shutil.which("git") and shutil.which("docker") and thisOS == "Linux":
                os.chdir(config.localStorage)
                print2("Setting up 'Perplexica' ...")
                try:
                    os.system(f"{shutil.which('git')} clone https://github.com/ItzCrazyKns/Perplexica.git")
                    os.chdir(perplexica_directory)
                    os.system(f"{shutil.which('docker')} compose up -d")
                    config.perplexica_directory = perplexica_directory
                    config.saveConfig()
                except:
                    print2("Failed setting up Perplexica! Read: https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Perplexica%20Integration.md for manual setup.")
                    webbrowser.open("https://github.com/eliranwong/toolmate/blob/main/package/toolmate/docs/Perplexica%20Integration.md")
        if config.perplexica_directory:
            if not isServerAlive(config.perplexica_ip, config.perplexica_port):
                os.chdir(config.perplexica_directory)
                os.system(f"{shutil.which('docker')} compose up -d")
            while not isServerAlive(config.perplexica_ip, config.perplexica_port):
                ...
            webbrowser.open(f"http://{config.perplexica_ip}:{config.perplexica_port}")
        else:
            QMessageBox.information(self.menu, "ToolMate AI", "Perplexica not found!")
        os.chdir(current_dir)


class TM(QApplication):
    
    def __init__(self, argv):
        super().__init__(argv)
        config.mainWindowHidden = False
        self.lastInsertedText = ""

    # Open Desktop Assistant on application activate
    def event(self, event):
        if event.type() == QEvent.ApplicationActivate and hasattr(config, "desktopAssistant") and (not config.desktopAssistant.isVisible() or not config.desktopAssistant.isActiveWindow()):
            config.desktopAssistant.show()
            clipboardText = pyperclip.paste()
            insertedText = f"# Context\n\n{clipboardText}" if clipboardText else ""
            if config.pasteTextOnWindowActivation and insertedText and not (insertedText == self.lastInsertedText) and not config.desktopAssistant.centralWidget.userInputMultiline.toPlainText().endswith(insertedText):
                config.desktopAssistant.centralWidget.userInputMultiline.insertPlainText(insertedText)
                self.lastInsertedText = insertedText
        """
#!/usr/bin/env bash

# work with text selection
# Install `xsel` and `xdotool`
# > apt install xsel xdotool
# To use this script, users need to:
# 1. Launch "Settings"
# 2. Go to "Keyboard" > "Keyboard Shortcuts" > "View and Customise Shortcuts" > "Custom Shortcuts"
# 3. Select "+" to add a custom shortcut and enter the following information, e.g.:

# Name: ToolMate AI
# Command: /home/username/toolmate/ToolMate
# Shortcut: Ctrl + Alt + L

selected_text=$(echo "$(xsel -o)" | sed 's/"/\"/g')
echo $selected_text | xsel --clipboard
toolmate=$(xdotool search --name "ToolMate AI")
xdotool windowmap $toolmate
xdotool windowactivate $toolmate
#xdotool type --window $toolmate $selected_text
        """
        return super().event(event)


def main():
    #app = QApplication(sys.argv)
    app = TM(sys.argv)
    app.setQuitOnLastWindowClosed(False)

    icon = QIcon(iconFile)
    trayIcon = ToolMateHub(icon)
    trayIcon.show()

    # unload llama.cpp model to free VRAM
    try:
        config.llamacppToolModel.close()
        print("Llama.cpp model unloaded!")
    except:
        pass

    sys.exit(app.exec())

if __name__ == "__main__":
    main()


import os
from toolmate import config, getAutogenConfigList, getAutogenCodeExecutionConfig, package, packageFolder
if not hasattr(config, "max_agents"):
    config.max_agents = 5
if not hasattr(config, "max_group_chat_round"):
    config.max_group_chat_round = 12
if not hasattr(config, "use_oai_assistant"):
    config.use_oai_assistant = False

from toolmate import print2, getCurrentModel

from autogen.agentchat.contrib.captainagent import CaptainAgent
from autogen import UserProxyAgent
from autogen import filter_config
import os, traceback, re, datetime, argparse
from pathlib import Path
from urllib.parse import quote
from toolmate.utils.prompts import Prompts
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.styles import Style
#from prompt_toolkit import PromptSession
#from prompt_toolkit.history import FileHistory

# Reference: https://microsoft.github.io/autogen/docs/reference/agentchat/contrib/agent_builder
class AutoCaptainAgent:

    def __init__(self):
        os.environ["BING_API_KEY"] = config.bing_api_key
        os.environ["RAPID_API_KEY"] = config.rapid_api_key

        self.llm = getCurrentModel()

        # prompt style
        self.promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        self.prompts = Prompts()

    def getSavePath(self, title=""):
        storageDir = os.path.join(os.path.expanduser('~'), package)
        if os.path.isdir(storageDir):
            folder = storageDir
        elif config.storagedirectory:
            folder = config.storagedirectory
        else:
            folder = os.path.join(packageFolder, "files")
        folder = os.path.join(folder, "autogen", "captain")
        if not os.path.isdir(folder):
            Path(folder).mkdir(parents=True, exist_ok=True)
        if title:
            title = "_" + quote(title, safe="")
        currentTime = re.sub(r"[\. :]", "_", str(datetime.datetime.now()))
        return os.path.join(folder, f"{currentTime}{title}.json")

    def getResponse(self, query):
        # https://ag2ai.github.io/ag2/blog/2024/11/15/CaptainAgent

        filter_dict = {"tags": [config.llmInterface]}
        config_list = filter_config(getAutogenConfigList(), filter_dict)

        llm_config={
            #"cache_seed": 42,  # seed for caching and reproducibility
            "config_list": config_list,  # a list of OpenAI API configurations
            "temperature": config.llmTemperature,  # temperature for sampling
            "timeout":  config.llm_timeout,
        }  # configuration for autogen's enhanced inference API which is compatible with OpenAI API

        # https://ag2ai.github.io/ag2/docs/topics/captainagent/configurations
        nested_config = {
            "autobuild_init_config": {
                "config_file_or_env": "OAI_CONFIG_LIST",
                "builder_model": self.llm,
                "agent_model": self.llm,
            },
            "autobuild_build_config": {
                "default_llm_config": llm_config,
                "code_execution_config": getAutogenCodeExecutionConfig(),
                "coding": True,
                "use_oai_assistant": config.use_oai_assistant,
                "library_path_or_json": os.path.join(packageFolder, "captainagent_expert_library.json"),
            },
            "autobuild_tool_config": {
                "tool_root": "default",  # this will use the tool library we provide
                "retriever": "paraphrase-multilingual-mpnet-base-v2",
            },
            "group_chat_config": {"max_round": config.max_group_chat_round},
            "group_chat_llm_config": llm_config.copy(),
        }

        ## build agents
        captain_agent = CaptainAgent(
            name="captain_agent",
            llm_config=llm_config,
            nested_config=nested_config,
            #agent_lib=os.path.join(packageFolder, "captainagent_expert_library.json"), # duplicate
            #tool_lib="default", # duplicate
            code_execution_config=getAutogenCodeExecutionConfig(),
            agent_config_save_path=None if config.code_execution_use_docker else self.getSavePath(),
        )
        user_proxy = UserProxyAgent(
            name="user_proxy",
            human_input_mode="NEVER"
        )
        chatResult = user_proxy.initiate_chat(captain_agent, message=query)
        return chatResult.chat_history

    def promptTask(self):
        self.print(f"<{config.terminalCommandEntryColor1}>Please specify a task below:</{config.terminalCommandEntryColor1}>")
        return self.prompts.simplePrompt(style=self.promptStyle)

    def promptConfig(self):
        print2("# AutoGen Configurations")
        self.print("Enter maximum number of agents:")
        max_agents = self.prompts.simplePrompt(numberOnly=True, style=self.promptStyle, default=str(config.max_agents),)
        if max_agents and int(max_agents) > 1:
            config.max_agents = int(max_agents)
        self.print("Enter maximum number of rounds in a group chat:")
        max_group_chat_round = self.prompts.simplePrompt(numberOnly=True, style=self.promptStyle, default=str(config.max_group_chat_round),)
        if max_group_chat_round and int(max_group_chat_round) > 1:
            config.max_group_chat_round = int(max_group_chat_round)
        self.print("Do you want to use OpenAI Assistant API (y/yes/N/NO)?")
        userInput = self.prompts.simplePrompt(style=self.promptStyle, default="y" if config.use_oai_assistant else "NO")
        if userInput:
            config.use_oai_assistant = True if userInput.strip().lower() in ("y", "yes") else False
        self.print("Do you want execute code in a docker image (y/yes/N/NO)?")
        userInput = self.prompts.simplePrompt(style=self.promptStyle, default="y" if config.code_execution_use_docker else "NO")
        if userInput:
            config.code_execution_use_docker = True if userInput.strip().lower() in ("y", "yes") else False
        if config.code_execution_use_docker:
            self.print("Please specify the docker image name below:")
            userInput = self.prompts.simplePrompt(style=self.promptStyle, default=config.code_execution_image)
            if userInput:
                config.code_execution_image = userInput
        self.print("Enter timeout for each code execution in seconds:")
        code_execution_timeout = self.prompts.simplePrompt(numberOnly=True, style=self.promptStyle, default=str(config.code_execution_timeout),)
        if code_execution_timeout and int(code_execution_timeout) > 1:
            config.code_execution_timeout = int(code_execution_timeout)
        config.saveConfig()

    def run(self):
        self.promptConfig()

        self.print(f"<{config.terminalCommandEntryColor1}>AutoGen Captain Agent launched!</{config.terminalCommandEntryColor1}>")
        self.print(f"""[press '{str(config.hotkey_exit).replace("'", "")[1:-1]}' to exit]""")
        while True:
            self.print(f"<{config.terminalCommandEntryColor1}>Hi! I am ready for a new task.</{config.terminalCommandEntryColor1}>")
            task = self.promptTask()
            if task == config.exit_entry:
                break
            try:
                self.getResponse(task)
            except:
                self.print(traceback.format_exc())
                break

        print2("\n\nAutoGen Captain Agent closed!")



    def print(self, message):
        #print(message)
        print_formatted_text(HTML(message))

def main():
    parser = argparse.ArgumentParser(description="AutoGen Captain Agent cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="execution task")
    parser.add_argument('-a', '--agents', action='store', dest='agents', type=int, help="maximum number of agents")
    parser.add_argument('-ed', '--executeindocker', action='store_true', dest='executeindocker', help="execute code in docker")
    parser.add_argument('-et', '--executiontimeout', action='store', dest='executiontimeout', type=int, help="timeout for each code execution in seconds")
    parser.add_argument('-oaia', '--oaiassistant', action='store_true', dest='oaiassistant', help="use OpenAI Assistant API")
    parser.add_argument('-r', '--rounds', action='store', dest='rounds', type=int, help="maximum round of group chat")
    # Parse arguments
    args = parser.parse_args()

    builder = AutoCaptainAgent()

    if args.agents and args.agents > 0:
        config.max_agents = args.agents

    if args.rounds and args.rounds > 0:
        config.max_group_chat_round = args.rounds

    if args.executiontimeout and args.executiontimeout > 0:
        config.code_execution_timeout = args.executiontimeout

    config.code_execution_use_docker = True if args.executeindocker else False
    config.use_oai_assistant = True if args.oaiassistant else False

    if args.default:
        builder.getResponse(task=args.default)
    else:
        builder.run()

if __name__ == '__main__':
    main()


"""
ToolMate AI Plugin - search Tavily

search Tavily

[TOOL_CALL]
"""

from toolmate import config

if not config.isLite and config.online:

    from toolmate import getTavilyClient

    def search_tavily(function_args):
        config.stopSpinning()
        if function_args:
            query = function_args.get("query")
            config.currentMessages[-1] = {"role": "user", "content": query}
        else:
            query = config.currentMessages[-1]["content"]
        context = getTavilyClient().get_search_context(query=query)
        return context

    functionSignature = {
        "examples": [
            "Search Tavily",
        ],
        "name": "search_tavily",
        "description": "Search for online information with Tavily",
        "parameters": {
            "type": "object",
            "properties": {} if not config.tool_selection_agent else {
                "query": {
                    "type": "string",
                    "description": "The original request in detail, including any supplementary information",
                },
            },
            "required": [] if not config.tool_selection_agent else ["query"],
        },
    }

    config.addFunctionCall(signature=functionSignature, method=search_tavily)
    config.inputSuggestions.append("Search Tavily: ")


"""
ToolMate AI Plugin - ask llama.cpp server

Ask llama.cpp server Model for conversation only; no function calling

[TOOL_CALL]
"""

from toolmate import config
from toolmate.utils.call_llamacppserver import CallLlamaCppServer

def llamacppserver(function_args):
    config.stopSpinning()
    if function_args:
        query = function_args.get("query") # required
        config.currentMessages[-1] = {"role": "user", "content": query}
    else:
        query = config.currentMessages[-1]["content"]
    completion = CallLlamaCppServer.regularCall(config.currentMessages)
    config.toolmate.streamCompletion(completion, openai=True)
    return ""

functionSignature = {
    "examples": [
        "Ask Llama.cpp Server",
    ],
    "name": "llamacppserver",
    "description": "Ask Llama.cpp Server to chat or provide information",
    "parameters": {
        "type": "object",
        "properties": {} if not config.tool_selection_agent else {
            "query": {
                "type": "string",
                "description": "The original request in detail, including any supplementary information",
            },
        },
        "required": [] if not config.tool_selection_agent else ["query"],
    },
}

config.addFunctionCall(signature=functionSignature, method=llamacppserver)
config.inputSuggestions.append("Ask Llama.cpp Server: ")


from toolmate import config, print2
from toolmate.api_client import getToolmate
from toolmate import getOllamaServerClient, exportOllamaModels, isServerAlive
from toolmate.utils.assistant import ToolMate
from prompt_toolkit.shortcuts import set_title, clear_title
from pathlib import Path
import os, shutil, argparse, re

def main():
    print(f"Setting up {config.toolMateAIName} ...")

    # Create the parser
    parser = argparse.ArgumentParser(description="ToolMate AI setup options")
    # Add arguments
    parser.add_argument('-ag', '--autogen', action='store_true', dest='autogen', help="configure AutoGen integration; applicable to AutoGen integrated tools")
    parser.add_argument('-b', '--backend', action='store_true', dest='backend', help="configure AI backend and models")
    parser.add_argument('-cs', '--chatsystem', action='store_true', dest='chatsystem', help="configure chat system message")
    parser.add_argument('-d', '--developer', action='store', dest='developer', help="configure developer mode; true / false")
    parser.add_argument('-ed', '--editor', action='store_true', dest='editor', help="configure custom editor")
    parser.add_argument('-ec', '--editconfigs', action='store_true', dest='editconfigs', help="configure config.py")
    parser.add_argument('-em', '--exportmodels', action='store', dest='exportmodels', help="""export models, downloaded with ollama, to ~/toolmate/LLMs/gguf/; pass a list of models for the export, e.g. "['llama3.2:1b','llama3.2:3b']"; pass an empty list "[]" to export all downloaded models""")
    parser.add_argument('-fb', '--fabric', action='store_true', dest='fabric', help="configure Fabric integration; applicable to Fabric integrated tools")
    parser.add_argument('-k', '--apikeys', action='store_true', dest='apikeys', help="configure API keys")
    parser.add_argument('-m', '--menu', action='store_true', dest='menu', help="setup menu")
    parser.add_argument('-mo', '--maximumoutput', action='store_true', dest='maximumoutput', help="configure maximum output tokens")
    parser.add_argument('-p', '--plugins', action='store_true', dest='plugins', help="configure plugins")
    parser.add_argument('-rt', '--riskthreshold', action='store_true', dest='riskthreshold', help="configure the risk threshold for user confirmation before code execution")
    parser.add_argument('-sg', '--speechgeneration', action='store_true', dest='speechgeneration', help="configure speech generation")
    parser.add_argument('-so', '--searchoptions', action='store_true', dest='searchoptions', help="configure search options")
    parser.add_argument('-sr', '--speechrecognition', action='store_true', dest='speechrecognition', help="configure speech recognition")
    parser.add_argument('-t', '--temperature', action='store_true', dest='temperature', help="configure inference temperature")
    parser.add_argument('-tms', '--tmsystems', action='store_true', dest='tmsystems', help="configure chat system messages for running with commands `tms1`, `tms2`, `tms3`, ... `tms20`")
    parser.add_argument('-tmt', '--tmtools', action='store_true', dest='tmtools', help="configure tools for running with commands `tmt1`, `tmt2`, `tmt3`, ... `tmt20`")
    parser.add_argument('-ta', '--toolagent', action='store_true', dest='toolagent', help="configure tool selection agent")
    parser.add_argument('-ws', '--windowsize', action='store_true', dest='windowsize', help="configure context window size")
    parser.add_argument('-ww', '--wordwrap', action='store', dest='wordwrap', help="configure word wrap; true / false")
    # Parse arguments
    args = parser.parse_args()
    # Check what kind of arguments were provided and perform actions accordingly

    # update to the latest version
    config.tempInterface = ""
    config.custom_config = ""
    config.initialCompletionCheck = False
    config.includeIpInDeviceInfoTemp = False
    config.defaultEntry = ""
    config.accept_default = False

    # set window title
    set_title(config.toolMateAIName)

    toolmate = ToolMate(plugins=True if args.tmsystems or args.tmtools or args.menu else False)

    if args.exportmodels:
        exportmodels = eval(args.exportmodels)
        if isinstance(exportmodels, list):
            exportOllamaModels(exportmodels)
        elif isinstance(exportmodels, str):
            exportOllamaModels(exportmodels.split())
        else:
            print2("""To export models, downloaded with ollama, either:
* pass a list of models for the export, e.g. "['llama3.2:1b','llama3.2:3b']"
* pass an empty list "[]" to export all downloaded models""")

    if args.menu:
        toolmate.runActions("...", setupOnly=True)
    if args.backend:
        toolmate.setLlmModel()
    if args.chatsystem:
        toolmate.setCustomSystemMessage()
    if args.developer:
        if args.developer.lower() == "true":
            config.developer = True
            print2("Developer mode enabled!")
        elif args.developer.lower() == "false":
            config.developer = False
            print2("Developer mode disabled!")
        else:
            print2("Developer mode unchanged! Accept 'True' or 'False' only!")
    if args.editconfigs:
        toolmate.editConfigs()
    if args.apikeys:
        toolmate.changeAPIkeys()
    if args.maximumoutput:
        toolmate.setMaxTokens()
    if args.plugins:
        toolmate.selectPlugins()
    if args.tmsystems:
        toolmate.setTmsMessages()
    if args.tmtools:
        toolmate.setTmtTools()
    if args.speechgeneration:
        toolmate.setTextToSpeechConfig()
    if args.speechrecognition:
        toolmate.setSpeechToTextConfig()
    if args.temperature:
        toolmate.setTemperature()
    if args.toolagent:
        toolmate.setToolSelectionConfigs()
    if args.riskthreshold:
        toolmate.manageCodeExecutionRisk()
    if args.windowsize:
        toolmate.setContextWindowSize()
    if args.wordwrap:
        if args.wordwrap.lower() == "true":
            config.wrapWords = True
            print2("Word wrap enabled!")
        elif args.wordwrap.lower() == "false":
            config.wrapWords = False
            print2("Word wrap disabled!")
        else:
            print2("Word wrap unchanged! Accept 'True' or 'False' only!")
    if args.searchoptions:
        toolmate.changeSearchSettings()
    if args.editor:
        toolmate.setCustomTextEditor()
    if args.autogen:
        toolmate.setAutoGenConfig()
    if args.fabric:
        toolmate.setFabricPatternsDirectory()

    # unload llama.cpp model to free VRAM
    try:
        config.llamacppToolModel.close()
        print("Llama.cpp model unloaded!")
    except:
        pass
    if config.llmInterface == "ollama":
        getOllamaServerClient().generate(model=config.ollamaToolModel, keep_alive=0, stream=False,)
        print(f"Ollama model '{config.ollamaToolModel}' unloaded!")
    if hasattr(config, "llamacppToolModel"):
        del config.llamacppToolModel

    # delete temporary content
    try:
        tempFolder = os.path.join(config.toolMateAIFolder, "temp")
        shutil.rmtree(tempFolder, ignore_errors=True)
        Path(tempFolder).mkdir(parents=True, exist_ok=True)
    except:
        pass

    # backup configurations
    config.saveConfig()

    host = config.toolmate_api_client_host
    port = config.toolmate_api_client_port
    if isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
        print("Reloading configurations ...")
        getToolmate({"reloadsettings": True})
        print("Reloaded!")

    # clear title
    clear_title()

if __name__ == "__main__":
    main()



import requests, argparse, json, sys, os, pprint, re, shutil, pydoc, pyperclip
from toolmate import config, convertOutputText, wrapText, startSpinning, stopSpinning, readTextFile, writeTextFile, print2, print3, getPygmentsStyle, showErrors, isServerAlive, getLlms, searchFolder, getCliOutput
from toolmate.utils.tts_utils import TTSUtil
from toolmate.utils.single_prompt import SinglePrompt

import pygments
from pygments.lexers.markup import MarkdownLexer
from pygments.lexers.python import PythonLexer
from prompt_toolkit.formatted_text import PygmentsTokens
from prompt_toolkit import print_formatted_text
from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from pathlib import Path


def highlightMarkdownSyntax(content):
    try:
        tokens = list(pygments.lex(content, lexer=MarkdownLexer()))
        print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
    except:
        print(content)

def highlightPythonSyntax(content, pformat=True):
    if pformat:
        try:
            content = pprint.pformat(content)
        except:
            pass
    try:
        tokens = list(pygments.lex(content, lexer=PythonLexer()))
        print_formatted_text(PygmentsTokens(tokens), style=getPygmentsStyle())
    except:
        print(content)

def configs():
    configFile = os.path.join(config.toolMateAIFolder, "config.py")
    print3(f"Reading: {configFile}")
    content = readTextFile(configFile)
    highlightPythonSyntax(content, pformat=False)
    print("")
    print2("```path")
    print(configFile)
    print2("```")

def chat():
    main(chat=True)

def cmd():
    main(defaultTool="command")

def task():
    main(defaultTool="task")

def python():
    main(defaultTool="execute_python_code")

def google():
    main(defaultTool="search_google")

def online():
    main(defaultTool="online")

def mp3():
    main(defaultTool="download_youtube_audio")

def mp4():
    main(defaultTool="download_youtube_video")

def reflection():
    main(defaultTool="reflection")

def deepReflection():
    main(defaultTool="deep_reflection")

def remember():
    main(defaultTool="save_memory")

def recall():
    main(defaultTool="search_memory")

def proxy():
    main(defaultTool="proxy")

def group():
    main(defaultTool="group")

def agents():
    main(defaultTool="agents")

def captain():
    main(defaultTool="captain")

def tms1():
    main(chatSystem=config.tms1)
def tms2():
    main(chatSystem=config.tms2)
def tms3():
    main(chatSystem=config.tms3)
def tms4():
    main(chatSystem=config.tms4)
def tms5():
    main(chatSystem=config.tms5)
def tms6():
    main(chatSystem=config.tms6)
def tms7():
    main(chatSystem=config.tms7)
def tms8():
    main(chatSystem=config.tms8)
def tms9():
    main(chatSystem=config.tms9)
def tms10():
    main(chatSystem=config.tms10)
def tms11():
    main(chatSystem=config.tms11)
def tms12():
    main(chatSystem=config.tms12)
def tms13():
    main(chatSystem=config.tms13)
def tms14():
    main(chatSystem=config.tms14)
def tms15():
    main(chatSystem=config.tms15)
def tms16():
    main(chatSystem=config.tms16)
def tms17():
    main(chatSystem=config.tms17)
def tms18():
    main(chatSystem=config.tms18)
def tms19():
    main(chatSystem=config.tms19)
def tms20():
    main(chatSystem=config.tms20)

def tmt1():
    main(chatSystem=config.tmt1)
def tmt2():
    main(chatSystem=config.tmt2)
def tmt3():
    main(chatSystem=config.tmt3)
def tmt4():
    main(chatSystem=config.tmt4)
def tmt5():
    main(chatSystem=config.tmt5)
def tmt6():
    main(chatSystem=config.tmt6)
def tmt7():
    main(chatSystem=config.tmt7)
def tmt8():
    main(chatSystem=config.tmt8)
def tmt9():
    main(chatSystem=config.tmt9)
def tmt10():
    main(chatSystem=config.tmt10)
def tmt11():
    main(chatSystem=config.tmt11)
def tmt12():
    main(chatSystem=config.tmt12)
def tmt13():
    main(chatSystem=config.tmt13)
def tmt14():
    main(chatSystem=config.tmt14)
def tmt15():
    main(chatSystem=config.tmt15)
def tmt16():
    main(chatSystem=config.tmt16)
def tmt17():
    main(chatSystem=config.tmt17)
def tmt18():
    main(chatSystem=config.tmt18)
def tmt19():
    main(chatSystem=config.tmt19)
def tmt20():
    main(chatSystem=config.tmt20)

def main(chat: bool = False, defaultTool=None, chatSystem=None, default=""):
    configFile = os.path.join(config.toolMateAIFolder, "config.py")
    # Create the parser
    parser = argparse.ArgumentParser(description = """ToolMate AI API client `tm` cli options;
                                    available shortcuts:
                                    `tmc` -> `tm -c`;
                                    `tmcmd` -> `tm -dt command`;
                                    `tmpython` -> `tm -dt execute_python_code`;
                                    `tmtask` -> `tm -dt task`;
                                    `tmgoogle` -> `tm -dt search_google` (internet connection required);
                                    `tmonline` -> `tm -dt online` (internet connection and SearXNG required);
                                    `tmmp3` -> `tm -dt download_youtube_audio` (internet connection required);
                                    `tmmp4` -> `tm -dt download_youtube_video` (internet connection required);
                                    `tmr` -> `tm -dt reflection`;
                                    `tmdr` -> `tm -dt deep_reflection`;
                                    `tmproxy` -> `tm -dt proxy` (full version only);
                                    `tmgroup` -> `tm -dt group` (full version only);
                                    `tmagents` -> `tm -dt agents` (full version only);
                                    `tmcaptain` -> `tm -dt captain` (full version only);
                                    `tmremember` -> `tm -dt save_memory` (full version only);
                                    `tmrecall` -> `tm -dt search_memory` (full version only);
                                    `tmt1` ... `tmt20` -> `tm -dt <custom_tool>` (determined by `config.tmt1` ... `config.tmt20`);
                                    `tms1` ... `tms20` -> `tm -cs <custom_chat_system_message>` (determined by `config.tms1` ... `config.tms20`, support pre-defined system messages or fabric patterns or custom entry);
                                    You may create your own aliases to make the shortcuts more memorable.""")
    # Add arguments
    parser.add_argument("default", nargs="*", default=None, help="instruction sent to ToolMate API server; work on previous conversation if not given.")
    parser.add_argument('-ar', '--autorag', action='store_true', dest='autorag', help="use AutoGen retriever for RAG tools, such as 'examine_files' and 'examine_web_content'; this feature is available in full version only")
    parser.add_argument('-b', '--backend', action='store', dest='backend', help="change AI backend if the model's backend is different")
    parser.add_argument('-bc', '--backupconversation', action='store_true', dest='backupconversation', help="back up the current conversation in ToolMate AI user directory")
    parser.add_argument('-bs', '--backupsettings', action='store_true', dest='backupsettings', help="back up the current settings in ToolMate AI user directory")
    parser.add_argument('-c', '--chat', action='store_true', dest='chat', help="enable to chat as an on-going conversation")
    parser.add_argument('-cf', '--chatfile', action='store', dest='chatfile', help="a chat file containing a saved conversation")
    parser.add_argument('-cp', '--chatpattern', action='store', dest='chatpattern', help=f"override chat system message for a single request, with a fabric pattern, in {config.fabricPatterns}; configure config.fabricPatterns to customise the path; use AI model assigned in ToolMate AI instead of in Fabric; this option cannot be used together with option 'chatsystem'; fabric is required to install separately")
    parser.add_argument('-cs', '--chatsystem', action='store', dest='chatsystem', help="override chat system message for a single request")
    parser.add_argument('-dt', '--defaulttool', action='store', dest='defaulttool', help="override default tool for a single request; applied when 'Tool Selection Agent' is disabled and no tool is specified in the request")
    parser.add_argument('-e', '--export', action='store', dest='export', help="export conversation; optionally used with -f option to specify a format for the export")
    parser.add_argument('-exec', '--execute', action='store_true', dest='execute', help="execute python code or system command; format a block of python code starting with '```python' or a block of system command starting with '```command'; ends the block with '```'")
    parser.add_argument('-f', '--format', action='store', dest='format', help="conversation output format; plain or chat; plain - readable format designed for sharing; chat - format for saving a conversation when used together with -e option; a saved chat file can be opened with -cf option; when this option is omitted, only the last assistant response is displayed by default")
    parser.add_argument('-ga', '--groupagents', action='store', dest='groupagents', type=int, help="group chat feature; maximum number of agents")
    parser.add_argument('-ged', '--groupexecuteindocker', action='store_true', dest='groupexecuteindocker', help="group chat feature; execute code in docker")
    parser.add_argument('-get', '--groupexecutiontimeout', action='store', dest='groupexecutiontimeout', type=int, help="group chat feature; timeout for each code execution in seconds")
    parser.add_argument('-goaia', '--groupoaiassistant', action='store_true', dest='groupoaiassistant', help="group chat feature; use OpenAI Assistant API; applicable to backend 'openai' only")
    parser.add_argument('-gr', '--grouprounds', action='store', dest='grouprounds', type=int, help="group chat feature; maximum number of rounds of discussion")
    parser.add_argument('-i', '--interactive', action='store_true', dest='interactive', help="interactive prompt, with auto-suggestions enabled, for writing instruction; do not use this option together with standard input or output")
    parser.add_argument('-imh', '--imageheight', action='store', dest='imageheight', type=int, help="image height; DALLE.3 supports 1024x1024 / 1024x1792 /1792x1024; Flux.1 natively supports any resolution up to 2 mp (1920x1088)")
    parser.add_argument('-imhd', '--imagehd', action='store_true', dest='imagehd', help="image quality in high definition")
    parser.add_argument('-ims', '--imagesteps', action='store', dest='imagesteps', type=int, help="image sampling steps")
    parser.add_argument('-imw', '--imagewidth', action='store', dest='imagewidth', type=int, help="image width; DALLE.3 supports 1024x1024 / 1024x1792 /1792x1024; Flux.1 natively supports any resolution up to 2 mp (1920x1088)")
    parser.add_argument('-info', '--information', action='store_true', dest='information', help="quick overview of server information")
    parser.add_argument('-ip', '--improveprompt', action='store_true', dest='improveprompt', help="toggle user prompt improvement for a single request")
    parser.add_argument('-k', '--key', action='store', dest='key', help="specify the API key for authenticating access to the ToolMate AI server")
    parser.add_argument('-m', '--model', action='store', dest='model', help="change AI model if the model is different")
    parser.add_argument('-ms', '--models', action='store_true', dest='models', help="show available AI backends and models")
    parser.add_argument('-md', '--markdown', action='store_true', dest='markdown', help="toggle markdown highlights of assistant response a single request")
    parser.add_argument('-mo', '--maximumoutput', action='store', dest='maximumoutput', type=int, help="override maximum output tokens for a single request; accepts non-negative integers; unaccepted values will be ignored")
    parser.add_argument('-p', '--port', action='store', dest='port', type=int, help="server port")
    parser.add_argument('-pa', '--paste', action='store_true', dest='paste', help="paste the clipboard text as a suffix to the instruction")
    parser.add_argument('-pd', '--powerdown', action='store_true', dest='powerdown', help="power down server")
    parser.add_argument('-py', '--copy', action='store_true', dest='copy', help="copy text output to the clipboard")
    parser.add_argument('-r', '--read', action='store_true', dest='read', help="read text output aloud")
    parser.add_argument('-rs', '--reloadsettings', action='store_true', dest='reloadsettings', help=f"Reload: 1. configurations in {configFile} 2. plugins")
    parser.add_argument('-rt', '--riskthreshold', action='store', dest='riskthreshold', type=int, help="risk threshold for user confirmation before code execution; 0 - always require confirmation; 1 - require confirmation only when risk level is medium or higher; 2 - require confirmation only when risk level is high or higher; 3 or higher - no confirmation required")
    parser.add_argument('-s', '--server', action='store', dest='server', help="server address; 'http://localhost' by default")
    parser.add_argument('-sd', '--showdescription', action='store_true', dest='showdescription', help="show description of the found items in search results; used together with option 'sc' or 'ss' or 'st'; show a fabric pattern content if used with option 'sp'")
    parser.add_argument('-sc', '--searchcontexts', action='store', dest='searchcontexts', help="search predefined contexts; use '@' to display all; use regex pattern to filter")
    parser.add_argument('-sp', '--searchpatterns', action='store', dest='searchpatterns', help=f"search fabric patterns in {config.fabricPatterns}; configure config.fabricPatterns to customise the search path; fabric is required to install separately")
    parser.add_argument('-ss', '--searchsystems', action='store', dest='searchsystems', help="search predefined system messages; use '@' to display all; use regex pattern to filter")
    parser.add_argument('-st', '--searchtools', action='store', dest='searchtools', help="search enabled tools; use '@' to display all; use regex pattern to filter")
    parser.add_argument('-t', '--temperature', action='store', dest='temperature', type=float, help="override inference temperature for a single request; accepted range: 0.0-2.0; unaccepted values will be ignored")
    parser.add_argument('-ta', '--toolagent', action='store_true', dest='toolagent', help="toggle tool selection agent for a single request")
    parser.add_argument('-vc', '--viewconfigs', action='store_true', dest='viewconfigs', help="view current server configurations")
    parser.add_argument('-wd', '--workingdirectory', action='store', dest='workingdirectory', help="working directory; current location by default")
    parser.add_argument('-ws', '--windowsize', action='store', dest='windowsize', type=int, help="override context window size for a single request; applicable to backends `llama.cpp` amd `ollama` only; accepts non-negative integers; unaccepted values will be ignored")
    parser.add_argument('-ww', '--wordwrap', action='store_true', dest='wordwrap', help="toggle word wrap of assistant response a single request")
    # Parse arguments
    args = parser.parse_args()

    def getPrefix(host, port):
        stopSpinning()

        endpoint = f"{host}:{port}/api/tools"
        url = f"""{endpoint}?query=@"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
            availableTools = json.loads(response.json())["results"].keys()
        except Exception as e:
            showErrors(e=e)
            return ""

        promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        instruction_history = os.path.join(historyFolder, "api_client")
        instruction_session = PromptSession(history=FileHistory(instruction_history))
        completer = FuzzyCompleter(WordCompleter(sorted([f"@{i}" for i in availableTools]), ignore_case=True))
        bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""
        instruction = SinglePrompt.run(style=promptStyle, promptSession=instruction_session, bottom_toolbar=bottom_toolbar, completer=completer)
        if instruction and not instruction.lower() == config.exit_entry:
            return instruction
        return ""

    mainOutput = ""
    if args.searchpatterns:
        if args.showdescription:
            # show content of a single pattern
            fabricPattern = os.path.join(os.path.expanduser(config.fabricPatterns), args.searchpatterns, "system.md")
            if os.path.isfile(fabricPattern):
                content = readTextFile(fabricPattern)
                highlightMarkdownSyntax(content)
            else:
                print3(f"File not found: {fabricPattern}")
        else:
            # search for a string in fabric pattern folder
            searchFolder(os.path.expanduser(config.fabricPatterns), args.searchpatterns, filter="system.md")
        return None
    host = args.server if args.server else config.toolmate_api_client_host
    port = args.port if args.port else config.toolmate_api_client_port
    if not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
        if (os.path.getsize(configFile) == 0 or not hasattr(config, "llmInterface") or not config.llmInterface) and shutil.which("tmsetup"):
            os.system("tmsetup")
        if shutil.which("nohup") and shutil.which("toolmateserver"):
            startSpinning()
            print2("Loading ToolMate AI ...")
            cli = f'''{shutil.which("nohup")} "{shutil.which("toolmateserver")}" > ~/toolmate/nohup-api-server.out 2>&1 &'''
            os.system(cli)
            # wait until the server is up
            while not isServerAlive(re.sub("^(http://|https://)", "", host, re.IGNORECASE), port):
                pass
            stopSpinning()
        else:
            print2("Failed to connect ToolMate AI! Run `toolmateserver` first!")

    if args.information or args.models or args.viewconfigs:

        #startSpinning()

        if args.information:
            query = "information"
        elif args.models:
            query = "models"
        elif args.viewconfigs:
            query = "configs"
        endpoint = f"{host}:{port}/api/status"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        #stopSpinning()
        
        try:
            results = response.json() if query == "configs" else json.loads(response.json())
            print(results) if query == "configs" else highlightPythonSyntax(results)
        except:
            print(response.text)

    elif args.searchtools is not None and args.searchtools.strip(): # -st given; search tools; ignore all other arguments
        startSpinning()

        query = args.searchtools.strip().lower()
        endpoint = f"{host}:{port}/api/tools"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription:
                for key, value in results.items():
                    print3(f"@{key}: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    elif args.searchsystems is not None and args.searchsystems.strip(): # -ss given; search predefined system messages; ignore all other arguments
        startSpinning()

        query = args.searchsystems.strip().lower()
        endpoint = f"{host}:{port}/api/systems"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription:
                for key, value in results.items():
                    print3(f"`{key}`: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    elif args.searchcontexts is not None and args.searchcontexts.strip(): # -sc given; search predefined contexts; ignore all other arguments
        startSpinning()

        query = args.searchcontexts.strip().lower()
        endpoint = f"{host}:{port}/api/contexts"

        url = f"""{endpoint}?query={query}"""
        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        try:
            response = requests.post(url, headers=headers)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()
        
        try:
            results = json.loads(response.json())["results"]
            if args.showdescription:
                for key, value in results.items():
                    print3(f"`{key}`: {value}")
            else:
                highlightPythonSyntax(list(results.keys()))
        except:
            print(response.text)

    else: # default given; "." for display current conversation only
        startSpinning()

        endpoint = f"{host}:{port}/api/toolmate"

        # formulate an instruction
        prefix = getPrefix(host, port) if args.interactive else ""
        cliDefault = " ".join(args.default) if args.default is not None else ""
        stdin_text = sys.stdin.read() if not sys.stdin.isatty() else ""
        if args.paste:
            clipboardText = getCliOutput("termux-clipboard-get") if config.terminalEnableTermuxAPI else pyperclip.paste()
        else:
            clipboardText = ""
        instruction = prefix + cliDefault + stdin_text + clipboardText + default
        if not instruction:
            # It simply uses the previously generated messages
            instruction = "."

        chatfile = args.chatfile if args.chatfile is not None and os.path.isfile(args.chatfile) else ""
        if chatfile or args.chat:
            chat = True
        if chatSystem is None and args.chatsystem:
            chatSystem = args.chatsystem
        
        # backend and model
        if args.backend and args.backend.lower() in getLlms().keys():
            backend = args.backend.lower()
        else:
            backend = None
        model = None
        if args.model:
            for b, ms in getLlms().items():
                if args.model in ms:
                    backend = b
                    model = args.model
                    break

        headers = {
            "Content-Type": "application/json",
            "X-API-Key": args.key if args.key else config.toolmate_api_client_key,
        }
        data = {
            "wd": args.workingdirectory if args.workingdirectory is not None and os.path.isdir(args.workingdirectory) else os.getcwd(),
            "backend": backend,
            "model": model,
            "instruction": instruction,
            "chat": chat,
            "chatfile": chatfile,
            "chatpattern": args.chatpattern,
            "chatsystem": chatSystem,
            "windowsize": args.windowsize,
            "maximumoutput": args.maximumoutput,
            "temperature": args.temperature,
            "defaulttool": defaultTool if defaultTool is not None else args.defaulttool,
            "toolagent": args.toolagent,
            "riskthreshold": args.riskthreshold,
            "execute": args.execute,
            "improveprompt": args.improveprompt,
            "autorag": args.autorag,
            "groupexecuteindocker": args.groupexecuteindocker,
            "groupexecutiontimeout": args.groupexecutiontimeout,
            "groupoaiassistant": args.groupoaiassistant,
            "groupagents": args.groupagents,
            "grouprounds": args.grouprounds,
            "imagehd": args.imagehd,
            "imageheight": args.imageheight,
            "imagewidth": args.imagewidth,
            "imagesteps": args.imagesteps,
            "markdown": args.markdown,
            "wordwrap": args.wordwrap,
            "backupconversation": args.backupconversation,
            "backupsettings": args.backupsettings,
            "reloadsettings": args.reloadsettings,
            "powerdown": args.powerdown,
        }
        try:
            response = requests.post(endpoint, headers=headers, json=data)
        except Exception as e:
            showErrors(e=e)
            stopSpinning()
            return None

        stopSpinning()

        if args.format and args.format.lower() in ("plain", "chat"):
            outputText = []
            if args.format.lower() == "plain":
                for i in json.loads(response.json()):
                    role = i.get("role", "")
                    content = i.get("content", "")
                    if role in ("user", "assistant"):
                        if role == "assistant":
                            content = convertOutputText(content.rstrip())
                        content = f"```{role}\n{content}\n```"
                        if args.export:
                            outputText.append(content)
                        else:
                            print(content)
            elif args.format.lower() == "chat":
                try:
                    output = json.loads(response.json())
                    if args.export:
                        outputText.append(pprint.pformat(output))
                    else:
                        pprint.pprint(output)
                except:
                    if args.export:
                        outputText.append(response.text)
                    else:
                        print(response.json())
            if args.export:
                try:
                    writeTextFile(args.export, "\n".join(outputText))
                except Exception as e:
                    showErrors(e=e)
        else:
            try:
                output = json.loads(response.json())[-1]["content"]
                output = convertOutputText(output)
                if args.export:
                    try:
                        writeTextFile(args.export, output)
                    except Exception as e:
                        showErrors(e=e)
                    return None
                if args.wordwrap:
                    current_wordwrap = config.wrapWords
                    config.wrapWords = not config.wrapWords
                if args.markdown:
                    current_toolmate_api_client_markdown = config.toolmate_api_client_markdown
                    config.toolmate_api_client_markdown = not config.toolmate_api_client_markdown
                outputContent = wrapText(output) if config.wrapWords else output
                highlightMarkdownSyntax(outputContent) if config.toolmate_api_client_markdown else print(outputContent)
                # restore configurations
                if args.wordwrap and not args.backupsettings:
                    config.wrapWords = current_wordwrap
                if args.markdown and not args.backupsettings:
                    config.toolmate_api_client_markdown = current_toolmate_api_client_markdown
                # copy response to clipboard
                if args.copy:
                    pydoc.pipepager(output, cmd="termux-clipboard-set") if config.terminalEnableTermuxAPI else pyperclip.copy(output)
                    print2(f"{config.divider}\nCopied!")
                # read aloud response
                if args.read:
                    TTSUtil.play(output)
                mainOutput = output
            except Exception as e:
                showErrors(e=e)
                print(response.text)
                mainOutput = response.text
    if default:
        return mainOutput

def getToolmate(data: dict):
    if not "wd" in data:
        data["wd"] = os.getcwd()
    host = config.toolmate_api_client_host
    port = config.toolmate_api_client_port
    endpoint = f"{host}:{port}/api/toolmate"
    headers = {
        "Content-Type": "application/json",
        "X-API-Key": config.toolmate_api_client_key,
    }
    try:
        return requests.post(endpoint, headers=headers, json=data)
    except Exception as e:
        showErrors(e=e)
        return f"Error: {e}"

if __name__ == '__main__':
    main()
